/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.components;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Type;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractNode;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.NodeClassLoader;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Set;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Component} nodes 
 * in definitions. This loader checks sub component names, loads the definitons
 * referenced by the component nodes (and the "extends" attribute), and merges
 * all these definitions into a single one (by following inheritance rules).
 */

public class ComponentLoader extends AbstractLoader {

  /**
   * TODO.
   */

  private Map NAME_ATTRIBUTES;

  private MergeClassLoader mergeClassLoader;
  
  // -------------------------------------------------------------------------
  // Constructor
  // -------------------------------------------------------------------------

  public ComponentLoader () {
    NAME_ATTRIBUTES = new HashMap();
    NAME_ATTRIBUTES.put("component", "name");
    NAME_ATTRIBUTES.put("interface", "name");
    NAME_ATTRIBUTES.put("binding", "from");
    NAME_ATTRIBUTES.put("attribute", "name");
    NAME_ATTRIBUTES.put("coordinates", "name");
    NAME_ATTRIBUTES.put("property", "name");
    mergeClassLoader = new MergeClassLoader(getClass().getClassLoader());
  }

  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public Definition load (final String name, final Map context) 
    throws ADLException 
  {
    return load(new ArrayList(), name, context);
  }
  
  public Definition load (
    final List loaded,
    final String name,
    final Map context) throws ADLException 
  {
    if (loaded.contains(name)) {
      throw new ADLException("Cycle in definition references: " + loaded, null);
    }
    List l = new ArrayList(loaded);
    l.add(name);
    Definition d = clientLoader.load(name, context);
    if (d instanceof ComponentDefinition) {
      ComponentDefinition container = (ComponentDefinition)d;
      normalizeComponentContainer(container);
      if (container.getExtends() != null) {
        try {
          d = (Definition)merge((Node)d, (Node)resolveDefinitions(l, container.getExtends(), context));
        } catch (ADLException e) {
          throw new ADLException(
            "Cannot load super definition(s)", (Node)container, e);
        }
        container = (ComponentDefinition)d;
        container.setExtends(null);
      }
      resolveComponentContainer(l, container, container, context);
    }
    return d;
  }

  // -------------------------------------------------------------------------
  // First pass: normalization
  // -------------------------------------------------------------------------

  public void normalizeComponentContainer (final ComponentContainer container) 
    throws ADLException
  {
    Set names = new HashSet();
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      String name = comp.getName();
      if (name == null) {
        throw new ADLException("Component name missing", (Node)comp);
      }
      if (names.contains(name)) {
        throw new ADLException(
          "Duplicated component name '" + name + "'", (Node)comp);
      } else {
        names.add(name);
      }
      normalizeComponentContainer(comp);
      if (name.indexOf('/') != -1) {
        container.removeComponent(comp);
        addComponent(container, comp);
      }
    }
  }

  public void addComponent (
    final ComponentContainer container, Component comp) throws ADLException
  {
    String name = comp.getName();
    int p = name.indexOf('/');
    if (p == -1) {
      Component c = getComponent(container, name);
      if (c != null) {
        container.removeComponent(c);
        comp = (Component)merge((Node)comp, (Node)c);
      }
      container.addComponent(comp);
    } else {
      Component parent = getComponent(container, name.substring(0, p));
      if (parent == null) {
        try {
          parent = (Component)comp.getClass().newInstance();
        } catch (InstantiationException e) {
          throw new Error("Internal error");
        } catch (IllegalAccessException e) {
          throw new Error("Internal error");
        }
        parent.setName(name.substring(0, p));
        container.addComponent(parent);
      }
      comp.setName(name.substring(p + 1));
      addComponent(parent, comp);
    }
  }

  public Component getComponent (
    final ComponentContainer container,
    final String name)
  {
    Component[] comps = container.getComponents();
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      if (comp.getName().equals(name)) {
        return comp;
      }
    }
    return null;
  }

  public Component getPathComponent (
    final ComponentContainer container,
    final String name)
  {
    int p = name.indexOf('/');
    if (p == -1) {
      return getComponent(container, name);
    } else {
      Component parent = getComponent(container, name.substring(0, p));
      return getPathComponent(parent, name.substring(p + 1));
    }
  }

  // -------------------------------------------------------------------------
  // Second pass: definition references resolution
  // -------------------------------------------------------------------------

  public void resolveComponentContainer (
    final List loaded,
    final ComponentContainer topLevelDefinition,
    final ComponentContainer container,
    final Map context) throws ADLException
  {
    Component[] comps = container.getComponents();
    Component[] newComps = new Component[comps.length];
    for (int i = 0; i < comps.length; i++) {
      newComps[i] = comps[i];
    }
    
    for (int i = 0; i < comps.length; i++) {
      Component comp = comps[i];
      resolveComponentContainer(loaded, topLevelDefinition, comp, context);
      String definition = comp.getDefinition();
      if (definition != null) {
        if (definition.indexOf('/') != -1) {
          // shared component
          for (int j = 0; j < comps.length; j++) {
            container.removeComponent(comps[j]);
          }
          for (int j = 0; j < comps.length; j++) {
            container.addComponent(newComps[j]);
            comps[j] = newComps[j];
          }
          newComps[i] = getPathComponent(topLevelDefinition, definition);
        } else {
          Definition d;
          try {
            d = resolveDefinitions(loaded, definition, context);
          } catch (ADLException e) {
            throw new ADLException(
              "Cannot load referenced definition(s)", (Node)comp, e);
          }
          newComps[i] = (Component)merge((Node)comp, (Node)d);
        }
        newComps[i].setDefinition(null);
      }
    }
    
    for (int i = 0; i < comps.length; i++) {
      container.removeComponent(comps[i]);
    }
    for (int i = 0; i < comps.length; i++) {
      container.addComponent(newComps[i]);
    }
  }

  public Definition resolveDefinitions (
    final List loaded,
    final String nameList, 
    final Map context) throws ADLException
  {
    int p = nameList.lastIndexOf(',');
    if (p == -1) {
      return load(loaded, nameList, context);
    } else {
      Definition d = resolveDefinitions(loaded, nameList.substring(0, p), context);
      Definition e = load(loaded, nameList.substring(p + 1), context);
      e = (Definition)clone((Node)e, new HashMap());
      e = (Definition)merge((Node)e, (Node)d);
      return e;
    }
  }

  public Node clone (final Node node, final Map clones) {
    Node copy;
    try {
      copy = node.astNewInstance();
    } catch (Exception e) {
      throw new Error("Internal error");
    }
    copy.astSetAttributes(node.astGetAttributes());
    copy.astSetDecorations(node.astGetDecorations());
    String[] types = node.astGetNodeTypes();
    for (int i = 0; i < types.length; ++i) {
      String type = types[i];
      Node[] subNodes = node.astGetNodes(type);
      for (int j = 0; j < subNodes.length; j++) {
        Node subNode = subNodes[j];
        Node subCopy;
        if (subNode == null) {
          subCopy = null;
        } else {
          subCopy = (Node)clones.get(subNode);
          if (subCopy == null) {
            subCopy = clone(subNode, clones);
            clones.put(subNode, subCopy);
          }
        }
        copy.astAddNode(type, subCopy);
      }
    }
    return copy;
  }

  // -------------------------------------------------------------------------
  // Utility methods: inheritance resolution
  // -------------------------------------------------------------------------

  public Node merge (Node elem, Node superElem) throws ADLException {
    Class ec = elem.getClass();
    Class sec = superElem.getClass();
    if (ec.getClassLoader() != sec.getClassLoader()) { // TODO ok?
      try {
        String code = Integer.toHexString(ec.hashCode()*sec.hashCode());
        String name = "org.objectweb.fractal.adl.merged.Merged" + code;
        Class merged = mergeClassLoader.mergeClass(
          name, elem.astGetType(), new Class[] { ec, sec });
        Node newElem = (Node) merged.newInstance();
        Node newSuperElem = (Node) merged.newInstance();
        shallowCopy(elem, newElem);
        shallowCopy(superElem, newSuperElem);
        elem = newElem;
        superElem = newSuperElem;
      } catch (Exception e) {
        throw new ADLException("Cannot merge AST classes", elem, e);
      }
    }
    
    // merges attributes
    Map attrs = elem.astGetAttributes();
    Map superAttrs = superElem.astGetAttributes();
    Iterator i = superAttrs.keySet().iterator();
    while (i.hasNext()) {
      String name = (String)i.next();
      String superValue = (String)superAttrs.get(name);
      String value = (String)attrs.get(name);
      if (superValue != null && value == null) {
        attrs.put(name, superValue);
      }
    }
    elem.astSetAttributes(attrs);
    
    // merges decorations
    Map decors = elem.astGetDecorations();
    Map superDecors = superElem.astGetDecorations();
    superDecors.putAll(decors);
    elem.astSetDecorations(superDecors);

    // merges sub nodes
    String[] superNodeTypes = superElem.astGetNodeTypes();
    for (int j = 0; j < superNodeTypes.length; ++j) {
      String superNodeType = superNodeTypes[j];
      Node[] superNodes = superElem.astGetNodes(superNodeType);
      Node[] nodes = elem.astGetNodes(superNodeType);

      List newNodes = new ArrayList();
      for (int k = 0; k < nodes.length; ++k) {
        newNodes.add(nodes[k]);
      }
      boolean changes = false;
      
      String nameAttr = (String)NAME_ATTRIBUTES.get(superNodeType);

      if (nameAttr == null) {
        if (superNodes.length > 0 && superNodes[0] != null) {
          if (nodes.length == 0 || nodes[0] == null) {
            //elem.astAddNode(superNodeType, superNodes[0]);
            newNodes.add(superNodes[0]);
            changes = true;
          } else {
            Node newN = merge(nodes[0], superNodes[0]);
            if (newN != nodes[0]) {
              newNodes.set(newNodes.indexOf(nodes[0]), newN);               
              changes = true;
            }
          }
        }
      } else {
        for (int k = 0; k < superNodes.length; ++k) {
          String superName = (String)superNodes[k].astGetAttributes().get(nameAttr);
          Node n = null;
          for (int l = 0; l < nodes.length; ++l) {
            String name = (String)nodes[l].astGetAttributes().get(nameAttr);
            if (name.equals(superName)) {
              n = nodes[l];
              break;
            }
          }
          if (n == null) {
            //elem.astAddNode(superNodeType, superNodes[k]);
            newNodes.add(superNodes[k]);
            changes = true;
          } else {
            Node newN = merge(n, superNodes[k]);
            if (newN != n) {
              newNodes.set(newNodes.indexOf(n), newN);
              changes = true;
            }
          }
        }
      }
      
      if (changes) {
        for (int k = 0; k < nodes.length; ++k) {
          elem.astRemoveNode(superNodeType, nodes[k]);
        }
        for (int k = 0; k < newNodes.size(); ++k) {
          elem.astAddNode(superNodeType, (Node)newNodes.get(k));
        }
      }
    }
    
    return elem;
  }
  
  private void shallowCopy (final Node src, final Node dst) {
    dst.astSetSource(src.astGetSource());
    dst.astSetAttributes(src.astGetAttributes());
    dst.astSetDecorations(src.astGetDecorations());
    String[] nodeTypes = src.astGetNodeTypes();
    for (int i = 0; i < nodeTypes.length; ++i) {
      Node[] srcNodes = src.astGetNodes(nodeTypes[i]);
      for (int j = 0; j < srcNodes.length; ++j) {
        dst.astAddNode(nodeTypes[i], srcNodes[j]);
      }
    }
  }
  
  static class MergeClassLoader extends NodeClassLoader {
            
    public MergeClassLoader (final ClassLoader parent) {
      super(parent);
    }
    
    public Class mergeClass (final String name, final String astNodeName, final Class[] classes)
      throws ClassNotFoundException 
    {
      try {
        return loadClass(name);
      } catch (ClassNotFoundException e) {        
      }
            
      Set itfs = new HashSet();
      for (int i = 0; i < classes.length; ++i) {
        Class[] citfs = classes[i].getInterfaces();
        for (int j = 0; j < citfs.length; ++j) {
          itfs.add(Type.getInternalName(citfs[j]));
        }
      }
      
      ClassWriter cw = generateClass(
        name, 
        astNodeName,
        Type.getInternalName(AbstractNode.class), 
        (String[])itfs.toArray(new String[itfs.size()]));
      
      /* DEBUG
      try {
        java.io.FileOutputStream fos =
          new java.io.FileOutputStream(name + ".class");
        fos.write(cw.toByteArray());
        fos.close();
      } catch (Exception _) {
      }
      */
      return defineClass(name, cw.toByteArray());
    }
  }
}
