/***
 * FractalGUI: a graphical tool to edit Fractal component configurations.
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Authors: Eric Bruneton, Patrice Fauvel
 */

package org.objectweb.fractal.gui.repository.lib;

import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.ParserException;
import org.objectweb.fractal.adl.XMLElement;

import org.objectweb.fractal.gui.model.ClientInterface;
import org.objectweb.fractal.gui.model.Component;
import org.objectweb.fractal.gui.model.Factory;
import org.objectweb.fractal.gui.model.Interface;
import org.objectweb.fractal.gui.model.ServerInterface;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * A converter from Fractal ADL definitions to {@link Component}s.
 */

class FractalAdlReader {

  /**
   * The loader used by this converter to load Fractal ADL definitions.
   */

  Loader loader;

  /**
   * The factory used by this converter to create the components.
   */

  Factory factory;

  /**
   * Converts the Fractal ADL template definition whose name is given.
   *
   * @param name the name of a Fractal ADL template definition.
   * @return a component converted from the given template definition.
   * @throws ParserException if something goes wrong.
   */

  public Component loadTemplate (final String name)
    throws ParserException
  {
    XMLElement desc;
    try {
      desc = loader.loadTemplateDescriptor(name);
    } catch (ParserException e) {
      throw new ParserException("Cannot load template '" + name + "'", e);
    }
    HashMap infos = new HashMap();
    // first pass: analyzes the dependencies between templates
    analyzeTemplate(name, desc, new HashMap(), infos);
    // second pass: creates the templates
    createTemplates(infos);
    // third pass: remove generated names
    Component c = ((TemplateInfo)infos.get(name)).value;
    cleanComponent(c);
    // return result
    return c;
  }

  /**
   * Analyzes the given template definition. This recursive method analyzes a
   * template definition and all its sub templates, and puts the collected
   * dependency information in the given hash map.
   *
   * @param path the path of the given template from the root template.
   * @param tmpl the template definition to be analyzed.
   * @param shared information that has been collected about shared components.
   *      This map contains the paths, from the root template, corresponding to
   *      the content of the &lt;sharing&gt; elements that have been analyzed.
   *      Each entry of this map associates a &lt;ref&gt; to a &lt;path&gt;.
   * @param infos information that has been collected about the templates that
   *      been analyzed. Each entry of this map associates a
   *      <tt>TemplateInfo</tt> to a template, designated by its path from the
   *      root template.
   * @throws ParserException if a problem occurs during the analysis.
   */

  private void analyzeTemplate (
    final String path,
    final XMLElement tmpl,
    final HashMap shared,
    final HashMap infos) throws ParserException
  {
    try {
      TemplateInfo info = new TemplateInfo(path);
      info.descriptor = tmpl;
      infos.put(path, info);
      if (tmpl.tag.equals("composite-template")) {
        XMLElement content = tmpl.getSubElement("composite-content");
        if (content == null) {
          return;
        }
        XMLElement sharing = content.getSubElement("sharing");
        if (sharing != null) {
          for (int i = 0; i < sharing.getSubElementCount(); ++i) {
            XMLElement elem = sharing.getSubElement(i);
            String epath = elem.getAttribute("path", null);
            String eref = elem.getAttribute("ref", null);
            shared.put(path + '/' + epath, path + '/' + eref);
          }
        }
        XMLElement comps = content.getSubElement("components");
        if (comps == null) {
          return;
        }
        info.requires = new String[comps.getSubElementCount()];
        for (int i = 0; i < info.requires.length; ++i) {
          XMLElement comp = comps.getSubElement(i);
          String name = path + '/' + comp.getAttribute("name", null);
          String ref = (String)shared.get(name);
          if (ref != null) {
            while (ref != null) {
              info.requires[i] = ref;
              ref = (String)shared.get(ref);
            }
          } else {
            info.requires[i] = name;
            String subName;
            try {
              subName = comp.getAttribute("implementation", null);
            } catch (ParserException e) {
              // optional component, whose implementation is not provided
              continue;
            }
            XMLElement subTmpl;
            try {
              subTmpl = loader.loadTemplateDescriptor(subName);
            } catch (ParserException e) {
              throw new ParserException(
                comp, "Cannot load template '" + subName + "'", e);
            }
            analyzeTemplate(name, subTmpl, shared, infos);
          }
        }
      }
    } catch (ParserException e) {
      throw new ParserException(tmpl, "Error while analyzing '" + path + "'",
        new ParserException(tmpl, e));
    }
  }

  /**
   * Creates the template components corresponding to the given definitions.
   * This method creates all the templates that do not have any dependencies,
   * then the templates that depend only on the previously instantiated
   * templates, and so on until all templates have been instantiated (or a
   * dependency cycle is detected, in which case an exception is thrown).
   *
   * @param infos information about the templates that must be created. Each
   *      entry of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @throws ParserException if a problem occurs during the creation of the
   *      template components.
   */

  private void createTemplates (final HashMap infos) throws ParserException {
    HashMap map = new HashMap(infos);
    while (map.size() > 0) {
      boolean ok = false;
      Iterator i = map.keySet().iterator();
      while (i.hasNext()) {
        String path = (String)i.next();
        TemplateInfo info = (TemplateInfo)map.get(path);
        if (info.value == null && info.instantiable(infos)) {
          createTemplate(info, path, infos);
          i.remove();
          ok = true;
        }
      }
      if (!ok) {
        throw new ParserException(
          "Dependency cycle due to shared components detected", null);
      }
    }
  }

  /**
   * Creates a template component corresponding to the given definition. This
   * method is only called when all the sub templates needed to create the
   * given template have already been created. These required sub templates
   * are retrieved from the given hash map.
   *
   * @param info information about the template component to be created.
   * @param path path of this template from the root template.
   * @param infos information about already created templates. Each entry
   *      of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @throws ParserException if a problem occurs during the creation of the
   *      template component.
   */

  private void createTemplate (
    final TemplateInfo info,
    final String path,
    final HashMap infos) throws ParserException
  {
    XMLElement tmpl = info.descriptor;
    try {
      if (tmpl.tag.equals("primitive-template")) {
        info.value = createPrimitiveTemplate(tmpl);
      } else if (tmpl.tag.equals("composite-template")) {
        info.value = createCompositeTemplate(info, infos);
      } else {
        info.value = createExternalTemplate(info);
      }
    } catch (ParserException e) {
      throw new ParserException(tmpl, "Error while creating '" + path + "'",
        new ParserException(tmpl, e));
    }
  }

  /**
   * Creates a primitive template corresponding to the given description.
   *
   * @param tmpl a &lt;primitive-template&gt; XML element.
   * @return a primitive template corresponding to the given description.
   * @throws ParserException if the template component cannot be created.
   */

  private Component createPrimitiveTemplate (final XMLElement tmpl)
    throws ParserException
  {
    String typeName = tmpl.getAttribute("implements", null);
    XMLElement type = loader.loadTypeDescriptor(typeName);
    XMLElement content = tmpl.getSubElement("primitive-content");
    XMLElement controller = tmpl.getSubElement("controller");
    String tmplDesc = "";
    String compDesc = "";
    XMLElement attrs = null;
    if (controller != null) {
      for (int i = 0; i < controller.getSubElementCount(); ++i) {
        XMLElement elem = controller.getSubElement(i);
        if (elem.tag.equals("attributes")) {
          attrs = elem;
        } else if (elem.tag.equals("template-controller")) {
          tmplDesc = elem.getAttribute("desc", null);
        } else if (elem.tag.equals("component-controller")) {
          compDesc = elem.getAttribute("desc", null);
        }
      }
    }
    Component c = factory.createComponent();
    loadComponentType(type, c);
    String name = tmpl.getAttribute("name", null);
    c.setType(name);
    if (content != null) {
      c.setImplementation(content.getAttribute("class", null));
    }
    c.setTemplateControllerDescriptor(tmplDesc);
    c.setComponentControllerDescriptor(compDesc);
    if (attrs != null) {
      readAttributes(c, attrs);
    }
    return c;
  }

  /**
   * Analyzes an &lt;attributes&gt; and sets these attributes on the given
   * component.
   *
   * @param c the component whose attributes must be initialized.
   * @param attrs an &lt;attributes&gt; element.
   * @throws ParserException if the attributes cannot be read.
   */

  private void readAttributes (final Component c, final XMLElement attrs)
    throws ParserException
  {
    c.setAttributeController(attrs.getAttribute("signature", ""));
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement attr = attrs.getSubElement(i);
      String name = attr.getAttribute("name", null);
      String value = attr.getAttribute("value", null);
      c.setAttribute(name, value);
    }
  }

  /**
   * Creates a composite template corresponding to the given definition. This
   * method is only called when all the sub templates needed to create the
   * given template have already been created. These required sub templates
   * are retrieved from the given hash map.
   *
   * @param info information about the template component to be created.
   * @param infos information about already created templates. Each entry
   *      of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @return the Component interface of the created template.
   * @throws ParserException if a problem occurs during the creation of the
   *      template component.
   */

  private Component createCompositeTemplate (
    final TemplateInfo info,
    final HashMap infos) throws ParserException
  {
    XMLElement tmpl = info.descriptor;
    String typeName = tmpl.getAttribute("implements", null);
    XMLElement type = loader.loadTypeDescriptor(typeName);
    XMLElement content = tmpl.getSubElement("composite-content");
    XMLElement controller = tmpl.getSubElement("controller");
    XMLElement components = null;
    XMLElement attrs = null;
    if (content != null) {
      components = content.getSubElement("components");
    }
    String tmplDesc = "";
    String compDesc = "";
    if (controller != null) {
      for (int i = 0; i < controller.getSubElementCount(); ++i) {
        XMLElement elem = controller.getSubElement(i);
        if (elem.tag.equals("attributes")) {
          attrs = elem;
        } else if (elem.tag.equals("template-controller")) {
          tmplDesc = elem.getAttribute("desc", null);
        } else if (elem.tag.equals("component-controller")) {
          compDesc = elem.getAttribute("desc", null);
        }
      }
    }
    // creates the composite template
    Component c = factory.createComponent();
    loadComponentType(type, c);
    String name = tmpl.getAttribute("name", null);
    c.setType(name);
    if (attrs != null) {
      readAttributes(c, attrs);
    }
    c.setTemplateControllerDescriptor(tmplDesc);
    c.setComponentControllerDescriptor(compDesc);

    HashMap subComps = new HashMap();
    subComps.put("this", c);
    int n = info.requires == null ? 0 : info.requires.length;
    for (int i = 0; i < n; ++i) {
      String path = info.requires[i];
      name = components.getSubElement(i).getAttribute("name", null);
      TemplateInfo tmplInfo = (TemplateInfo)infos.get(path);
      if (tmplInfo == null) {
        // optional component, whose implementation is not provided
        continue;
      }
      Component id = tmplInfo.value;
      if (!path.equals(info.path + '/'  + name)) {
        // shared component
        id = factory.createComponent(id);
      }
      id.setName(name);
      subComps.put(name, id);
      c.addSubComponent(id);
    }

    // creates the bindings between the sub template components
    if (content == null) {
      return c;
    }
    XMLElement bindings = content.getSubElement("bindings");
    if (bindings == null) {
      return c;
    }
    for (int i = 0; i < bindings.getSubElementCount(); ++i) {
      XMLElement bind = bindings.getSubElement(i);
      String clt = bind.getAttribute("client", null);
      int p = clt.indexOf('.');
      String cltComp = clt.substring(0, p);
      String cltItf = clt.substring(p + 1);
      String srv = bind.getAttribute("server", null);
      p = srv.indexOf('.');
      String srvComp = srv.substring(0, p);
      String srvItf = srv.substring(p + 1);

      Component client = (Component)subComps.get(cltComp);
      Component server = (Component)subComps.get(srvComp);

      if (client == null || server == null) {
        // one of the component is an optional component which is not provided:
        // skips this binding
        continue;
      }
      if (client == null) {
        throw new ParserException(
          bind, "No such sub template: " + cltComp, null);
      }
      if (server == null) {
        throw new ParserException(
          bind, "No such sub template: " + srvComp, null);
      }

      String cltItfPrefix = getItfPrefix(client, cltItf);
      if (cltItfPrefix == null) {
        throw new ParserException(bind, "No such client interface: "+clt, null);
      }
      Interface cItf = client.getClientInterface(cltItfPrefix);
      if (cItf == null && client.isComposite()) {
        cItf = client.getServerInterface(cltItfPrefix);
        if (cItf != null) {
          cItf = cItf.getComplementaryInterface();
        }
      }
      if (cItf == null) {
        throw new ParserException(bind, "No such client interface: "+clt, null);
      }

      Interface sItf = server.getServerInterface(srvItf);
      if (sItf == null && server.isComposite()) {
        sItf = server.getClientInterface(srvItf);
        if (sItf != null) {
          sItf = sItf.getComplementaryInterface();
        }
      }
      if (sItf == null) {
        throw new ParserException(bind, "No such server interface: "+srv, null);
      }

      client.bind((ClientInterface)cItf, (ServerInterface)sItf);
    }
    return c;
  }

  /**
   * Creates an external template corresponding to the given description.
   *
   * @param info information about the template component to be created.
   * @return a template corresponding to the given description.
   * @throws ParserException if the template component cannot be created.
   */

  private Component createExternalTemplate (final TemplateInfo info)
    throws ParserException
  {
    XMLElement tmpl = info.descriptor;
    String typeName = tmpl.getAttribute("implements", null);
    XMLElement type = loader.loadTypeDescriptor(typeName);
    Component c = factory.createComponent();
    loadComponentType(type, c);
    c.setName("<external>");
    return c;
  }

  /**
   * Creates the component type corresponding to the given description.
   *
   * @param node a &lt;component-type&gt; XML element.
   * @param component the component whose type must be created.
   * @throws ParserException if the component type cannot be created.
   */

  private void loadComponentType (
    final XMLElement node,
    final Component component) throws ParserException
  {
    for (int i = 0; i < node.getSubElementCount(); ++i) {
      XMLElement subElement = node.getSubElement(i);
      loadInterfaceTypes(subElement, component);
    }
  }

  /**
   * Creates the interface types corresponding to the given descriptions.
   *
   * @param node a &lt;provides&gt; or &lt;requires&gt; XML element.
   * @param component the component whose type must be created.
   * @throws ParserException if an interface type cannot be create.
   */

  private void loadInterfaceTypes (
    final XMLElement node,
    final Component component) throws ParserException
  {
    boolean server = node.tag.equals("provides");
    for (int i = 0; i < node.getSubElementCount(); ++i) {
      XMLElement subElement = node.getSubElement(i);
      String prefix = subElement.getAttribute("name", null);
      String type = subElement.getAttribute("signature", null);
      boolean mandatory =
        subElement.getAttribute("contingency", "mandatory").equals("mandatory");
      boolean singleton =
        subElement.getAttribute("cardinality", "single").equals("single");
      if (!singleton) {
        mandatory = false;
      }
      Interface itf;
      if (server) {
        itf = factory.createServerInterface(component);
      } else {
        itf = factory.createClientInterface(component);
      }
      itf.setName(prefix);
      itf.setSignature(type);
      itf.setIsOptional(!mandatory);
      itf.setIsCollection(!singleton);
      if (server) {
        component.addServerInterface((ServerInterface)itf);
      } else {
        component.addClientInterface((ClientInterface)itf);
      }
    }
  }

  /**
   * Returns the prefix name of the given interface name.
   *
   * @param c a component.
   * @param itfName the name of an interface of this component.
   * @return the prefix name corresponding to the given name.
   */

  private String getItfPrefix (final Component c, final String itfName) {
    for (int i = 0; i < 2; ++i) {
      List l = i == 0 ? c.getClientInterfaces() : c.getServerInterfaces();
      for (int j = 0; j < l.size(); ++j) {
        Interface itf = (Interface)l.get(j);
        String name = itf.getName();
        if (itf.isCollection()) {
          if (itf.getMasterCollectionInterface() == null &&
              itfName.startsWith(name))
          {
            return name;
          }
        } else if (itfName.equals(name)) {
          return name;
        }
      }
    }
    return null;
  }

  /**
   * Information about a template.
   */

  static class TemplateInfo {

    /**
     * The path of this template.
     */

    String path;

    /**
     * A template definition.
     */

    XMLElement descriptor;

    /**
     * The paths of the sub templates that are required to create the template
     * defined by <tt>descriptor</tt>. May be <tt>null</tt> if there are no such
     * sub templates.
     */

    String[] requires;

    /**
     * The component corresponding to <tt>descriptor</tt>. May be
     * <tt>null</tt> if this component has not been created yet.
     */

    Component value;

    /**
     * Constructs a new {@link FractalAdlReader.TemplateInfo} object.
     *
     * @param path the path of the template described by this object.
     */

    public TemplateInfo (final String path) {
      this.path = path;
    }

    /**
     * Tests if the template defined by <tt>descriptor</tt> can be instantiated.
     *
     * @param infos information about already instantiated templates. Each entry
     *      of this map associates a <tt>TemplateInfo</tt> to a template,
     *      defined by its path from the root template.
     * @return <tt>true</tt> if the template defined by <tt>descriptor</tt> can
     *      be instantiated, i.e., if all the templates it requires have already
     *      been instantiated (i.e., if the <tt>value</tt> field of their
     *      associated <tt>TemplateInfo</tt> object is not <tt>null</tt>).
     */

    public boolean instantiable (final HashMap infos) {
      if (requires != null) {
        for (int i = 0; i < requires.length; ++i) {
          TemplateInfo tmplInfo = (TemplateInfo)infos.get(requires[i]);
          if (tmplInfo != null && tmplInfo.value == null) {
            return false;
          }
        }
      }
      return true;
    }
  }

  /**
   * Removes the generated names from the given component. A component and its
   * interfaces may have names generated by the inverse converter, if some names
   * where missing in a component when it was stored.
   *
   * @param c the component to be cleaned.
   */

  private void cleanComponent (final Component c) {
    if (c.getName().startsWith("GENERATED-")) {
      c.setName("");
    }
    if (c.getType().startsWith("org.objectweb.fractal.gui.tmp.C")) {
      c.setType("");
    }
    List l = c.getClientInterfaces();
    for (int i = 0; i < l.size(); ++i) {
      Interface itf = (Interface)l.get(i);
      if (itf.getName().startsWith("GENERATED")) {
        itf.setName("");
      }
    }
    l = c.getServerInterfaces();
    for (int i = 0; i < l.size(); ++i) {
      Interface itf = (Interface)l.get(i);
      if (itf.getName().startsWith("GENERATED")) {
        itf.setName("");
      }
    }
    l = c.getSubComponents();
    for (int i = 0; i < l.size(); ++i) {
      cleanComponent((Component)l.get(i));
    }
  }
}
